#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/pgmspace.h>


#if defined(__AVR_AT90S2313__)
#  define OC1 PB3
#  define OCR OCR1
#  define DDROC DDRB
#elif defined(__AVR_AT90S2333__) || defined(__AVR_AT90S4433__)
#  define OC1 PB1
#  define DDROC DDRB
#  define OCR OCR1
#elif defined(__AVR_AT90S4414__) || defined(__AVR_AT90S8515__) || \
      defined(__AVR_AT90S4434__) || defined(__AVR_AT90S8535__) || \
      defined(__AVR_ATmega163__)
#  define OC1 PD5
#  define DDROC DDRD
#  define OCR OCR1A
#else
#  error "Don't know what kind of MCU you are compiling for"
#endif

#if defined(COM11)
#  define XCOM11 COM11
#elif defined(COM1A1)
#  define XCOM11 COM1A1
#else
#  error "need either COM1A1 or COM11"
#endif

#define F_CPU 4000000                           // Zum Beispiel 4Mhz-Quarz
#define UART_BAUD_RATE 9600                     // Wir versuchen mal mit 9600 Baud


#define DCFLowMin   6
#define DCFLowMax   12
#define DCFHighMin  16
#define DCFHighMax  22
#define DCFStartMin 160
#define DCFStartMax 200
//#define ITCSec      100


#define DCFYear		aDCFTime[5]
#define DCFMonth	aDCFTime[4]
#define DCFDay		aDCFTime[3]
#define DCFHour		aDCFTime[2]
#define DCFMin		aDCFTime[1]
#define DCFSec 	  	aDCFTime[0]		// DCF DateTime

#define ITCYear		aITCTime[5]
#define ITCMonth	aITCTime[4]
#define ITCDay		aITCTime[3]
#define ITCHour		aITCTime[2]
#define ITCMin		aITCTime[1]
#define ITCSec 	  	aITCTime[0]		// internal clock DateTime

#define EVTYear		aEVTTime[5]
#define EVTMonth	aEVTTime[4]
#define EVTDay		aEVTTime[3]
#define EVTHour		aEVTTime[2]
#define EVTMin		aEVTTime[1]
#define EVTSec 	  	aEVTTime[0]		// event DateTime

#define SecPuls 		0
#define EventFlag		1

#define ErrParity       0
#define ErrDCFHL        1
#define ErrDCFStart     2

const uint8_t aDCFValues[] PROGMEM ={0x1,0x2,0x4,0x8,0xA,0x14,0x28,0x0,0x1,0x2,0x4,0x8,0xA,0x14,0x0,0x1,0x2,0x4,0x8,0xA,0x14,0x1,0x2,0x4,0x1,0x2,0x4,0x8,0xA,0x1,0x2,0x4,0x8,0xA,0x14,0x28,0x50,0x0};

volatile uint8_t nCount;
volatile uint8_t nCntItc;
volatile uint8_t nCntPuls;
volatile uint8_t nPulsLength;
volatile uint8_t nDCFSyncCnt;
volatile uint8_t nSigLength;
volatile uint8_t nSigEval;
volatile uint8_t nCtrlFlags;
volatile uint8_t nValue;
volatile uint8_t nLastSigEval;
volatile uint8_t nErrorFlags;
volatile uint8_t nParity;

uint8_t aDCFTime[6];
uint8_t aITCTime[6];
uint8_t aEVTTime[6]={255,255,255,255,255,255};

SIGNAL(SIG_INTERRUPT1)
{    
    uint8_t nMCUCR = inb(MCUCR);

    nSigLength=nCount;	    
    
    if (!(nMCUCR & (1<<ISC10))) {
        nMCUCR |= (1<<ISC10);    	   /* raise int on rising edge */	
        PORTB &= ~1;        
    }
    else {
    	nMCUCR &= ~(1<<ISC10);              /* raise int on falling edge */
        PORTB |= 1;
    }
    
    if (nSigEval==2) 
      nSigEval=0;
    
    
    nSigEval++;    
	
    nCount=0;    
    
    outp(nMCUCR, MCUCR);    
        
}


SIGNAL (SIG_OUTPUT_COMPARE1A)
{
    nCount++;
    nCntItc++;  
        
}

void UartWrite (uint8_t nValue)
{
    outp(nValue, UDR);
    
    while(!(inb (USR) & (1 << UDRE))) {
       
    }
} 

uint8_t UartRead()
{
    while (!(inb (USR) & (1 << RXC))) {}
	  
    return(inb(UDR));    
}

uint8_t GetDCFBitValue(uint8_t nBitNo)
{
    return(PRG_RDB(&aDCFValues[nBitNo]));
}


void InternalClock(void)
{
        
    if (++ITCSec==60)        			//keep track of time, date, month, and year
    {
        ITCSec=0;
        if (++ITCMin==60) 
        {
            ITCMin=0;
            if (++ITCHour==24)
            {
                ITCHour=0;
                if (++ITCDay==32)
                {
                    ITCMonth++;
                    ITCDay=1;
                }
                else if (ITCDay==31) 
                {                    
                    if ((ITCMonth==4 || 6 || 9 || 11)) 
                    {
                        ITCMonth++;
                        ITCDay=1;
                    }
                }
                else if (ITCDay==30)
                {
                    if(ITCMonth==2)
                    {
                       ITCMonth++;
                       ITCDay=1;
                    }
                }              
                else if (ITCDay==29) 
                {
                    if((ITCMonth==2) && (ITCYear%4))
                    {
                        ITCMonth++;
                        ITCDay=1;
                    }                
                }                          
                if (ITCMonth==13)
                {
                    ITCMonth=1;
                    ITCYear++;
                }
            }
        }
    }  
}

void DCFBitReceived(uint8_t nBit)
{    
	if (DCFSec>=21) {  
      if (nBit & 1) {	  
	    nValue+=GetDCFBitValue(DCFSec-21);
	  }	
      nParity^=nBit;	  
	}	
	
    switch (DCFSec) {
	  case 21: 
      case 29: 
      case 36:
	    nParity=(nBit & 1);
		break;
      case 27: 
        DCFMin=nValue;
        nValue=0;  		
		break;	
	  case 28:
	  case 35:
	  case 58:
	    DCFYear=nValue;
        nValue=0;    
	    if (nParity){
	      nErrorFlags|=(1<<ErrParity);
		  nParity=0;
	    }  
		break;
      case 34:
        DCFHour = nValue;
        nValue=0;		
      case 41:
        DCFDay=nValue;
        nValue=0;
		break;	
      case 44:
        //DCFWDay = nValue;
        nValue=0;
		break;	
      case 49:
        DCFMonth=nValue;
        nValue=0;
		break;	 
    }
}


void SendDateTime(uint8_t aDT[], uint8_t nStart, uint8_t nStopp)
{

	uint8_t cId;
	
	UartWrite(nStart);
	for (cId=5;cId!=255;cId--) 
		UartWrite(aDT[cId]);
		
    UartWrite(nStopp);
	
}

void ReadDateTime(uint8_t aDT[])
{
	uint8_t cId;
	
	for (cId=5;cId!=255;cId--) {
	  /*while (!(inb (USR) & (1 << RXC))) {
	  }
      aDT[cId] = inb (UDR);        		*/
      aDT[cId]=UartRead();
	}
	  
}

uint8_t CompareTime(uint8_t nDT1[], uint8_t nDT2[])
{
	uint8_t cId,nRet;
	
	nRet=1;
	
	for (cId=5;cId!=255;cId--) {
	  if ((nDT1[cId] >= nDT2[cId]) & (nRet==1)) {
	    nRet=1;
	  }	
	  else {
	    nRet=0;
		break;
	  }
	}
	return(nRet);		
}


uint8_t main (void)
{
    uint8_t chUart;       	
				 
    outp(0xFF,DDRB);		                        /* portb for output */
    outp(0xFD,PORTB);			                    /* clear all */	
    
    outp(0x00, DDRD);			                    /* portd for input */       
    outp(0x08,PORTD);                               /* activate pullup on PD3 */
    
    outp((1<<ISC11), MCUCR);                        /* raise int1 on falling edge */
    outp((1<<INT1), GIMSK);                         /* enable external int0, int1 */
    
    outp(0,TIMSK);                                  /* disable timer interrupts */
    outp(0,TCCR1A);			                        /* Timer/Counter Control Register A Timer 1 */

    outp((1<<CS11)|(1<<CS10)|(1<<CTC1),TCCR1B);     /* Vorzaehler  */
    outp((1<<OCIE1A),TIMSK);	
    outw(OCR,0x271);                                //(F_CPU * 10) / (1000 * 64) = Timer1 fr 10ms berechnen
      
    outp((1 << RXEN) | (1 << TXEN), UCR);
    outp(F_CPU / (UART_BAUD_RATE * 16L) - 1, UBRR);       
    
	nDCFSyncCnt=3;
	nPulsLength=3;
        
    sei(); 
    	       
    for (;;) { 	  
	
        if (nCntItc>=100){      // internal clock  100 * 10ms = 1s
            InternalClock();
            nCntItc=0;
            nCtrlFlags|=(1<<SecPuls);		
        }
    	  
        if (nSigEval != nLastSigEval) {	      
            switch(nSigEval) {
            	case 1:                                                                 	  
            	    if ((nSigLength>DCFLowMin) & (nSigLength<DCFLowMax)) {
            		  DCFBitReceived(0);            
            		}
            		else { 
                		if ((nSigLength>DCFHighMin) & (nSigLength<DCFHighMax)) {
                		    DCFBitReceived(1);                     
            		    }
            		    else {
            		        nErrorFlags|=(1<<ErrDCFHL);
            		    }
            	    }
                    DCFSec++;                                 			      			
                    break;
            	  
            	case 2:   
                    if ((nSigLength>=DCFStartMin) & (nSigLength<=DCFStartMax)) {
            	        nErrorFlags=0;				
            		    DCFSec=0;
            		        			
                		if (DCFMin==0) 	// interne Uhr nach neuer Stunde Synchronisieren
                		    nDCFSyncCnt=3;
            			
                		if (nDCFSyncCnt>0)
                		    nDCFSyncCnt--;	
            		  
                		if ((nDCFSyncCnt==1) & (nErrorFlags==0)) {                 		  
                            ITCSec=DCFSec;   			
                            ITCMin=DCFMin;
                            ITCHour=DCFHour;                                     
                            ITCDay=DCFDay;   
                            ITCMonth=DCFMonth;
                            ITCYear=DCFYear;		      	      
                		}	      				
            	    }	              	   	      	  
            	    break;  
            }           
        	  
            nLastSigEval=nSigEval;
        }	  
    		  
        if ((nErrorFlags>0) & (nDCFSyncCnt>0))
            nDCFSyncCnt=3;	
         
         
        // UART
          
        chUart=0;
        
        if (inb (USR) & (1 << RXC)) {
            chUart = inb (UDR);        		
            switch (chUart) {
                case 200:		   
                    SendDateTime(aDCFTime,200,201);
                    break;
                case 202:		   
                    SendDateTime(aITCTime,202,203);
                    break;
                case 204:		   
                    SendDateTime(aEVTTime,204,205);
                    break;            
                case 206:
                    ReadDateTime(aEVTTime);
                    nCtrlFlags&=~(1<<EventFlag);
                    nCntPuls=0;
                    UartWrite(207);
                    break;            
                case 208:
                    nPulsLength=UartRead();;
                    UartWrite(209);
                    break;
                case 210:
                    UartWrite(nDCFSyncCnt);		                      
                    UartWrite(nErrorFlags);
                    UartWrite(211);
                    break;	
            }	
        }
    	  
    	// Ereignis 1
    	  	  
        if ((CompareTime(aITCTime,aEVTTime) == 1) & (!(nCtrlFlags & (1<<EventFlag)))){
            nCtrlFlags|=(1<<EventFlag);
            nCntPuls=nPulsLength;
            //PORTB&=~2;            
            PORTB|=2;    
        }    
    
        if ((nCntPuls>0) && (nCtrlFlags & (1<<SecPuls))) {	        
            nCntPuls--;		
            if (nCntPuls==0)		  
            //PORTB|=2;    
            PORTB&=~2;            
        }        
           
        nCtrlFlags&=~(1<<SecPuls);
        		
    }
      
        
    return (0);
}
